function [hqem_parameter,hqem_error] = hqem(sysbio_models,sysbio_model_variant,fitting_value,search_option)
% [hqem_parameter,hqem_error] = hqem(sysbio_models,sysbio_model_variant,fitting_value,search_option)
% ------------------------------- INPUT -------------------------------
% sysbio_models        : struct of SimBiology.Model objects
%       e.g. sysbio_models.m1, sysbio_models.m2, ...
%            m1, m2 et al. are Matlab SimBiology.Model objects.
% sysbio_model_variant : The variant object of systems biology model, which
%       contain the setting for parameter tuning.
% fitting_value        : struct of data for fitting
%       e.g. fitting_value.m1.target_species_name,
%            fitting_value.m1.time_span,
%            fitting_value.m1.target_species_value,
%            fitting_value.m1.target_species_value_scale, ...
%       target_species_name  : s-by-1
%       time_span : n-by-1 vector representing the time span of
%            the target data, e.g. [0;10;20]. (Note: use semicolon (;))
%       target_species_value : n-by-s, targetted species value for fitting.
%       target_species_value_scale  : 1, [], 1-by-s, n-by-1, or n-by-s.
% ------------------------------ OUTPUT -------------------------------
% hqem_parameter     : The obtained parameter sets
% hqem_error         : The fitting error caculated using the parameter sets

%% Source code
[m_sbioM_number, fitting_value] = jz_sbioEnsemCheckVars( sysbio_models, fitting_value );
[global_search_option, local_search_option] = jz_sbioEnsemCheckOption(search_option);
[m_sbioModels.m1, t_ans, global_search_option.LowerBound, global_search_option.UpperBound] = jz_sbioEnsemParamModel( sysbio_models.m1, sysbio_model_variant, 'haltonset', 1, [global_search_option.LowerBound; global_search_option.UpperBound] );
for t_m = 2:m_sbioM_number
    m_sbioModels.(['m' num2str(t_m)]) = jz_sbioEnsemParamModel( sysbio_models.(['m' num2str(t_m)]), sysbio_model_variant, 'none', 1, [] );
end
save(global_search_option.IntermediateFileName, 'm_sbioModels', 'global_search_option', 'local_search_option', 'fitting_value');

jz_sbioEnsemDisplay(m_sbioModels, fitting_value)

if global_search_option.MatlabPoolSize > 1
    if matlabpool('size') > 1
        matlabpool('close');
    end
    matlabpool('open',global_search_option.MatlabPoolSize);
end

if strcmp(global_search_option.UseParallel,'yes')
    sbioproject = ['jzsbioproj_', upper(num2hex( sum([1000*clock rand(1,1)]) ))];
    cmd_savesbioproj = ['sbiosaveproject(''' sbioproject ''''];
    for t_m = 1:m_sbioM_number
        eval(['m' num2str(t_m) ' = m_sbioModels.([''m'' num2str(t_m)]);']);
        cmd_savesbioproj = [cmd_savesbioproj ',''m' num2str(t_m) ''''];
    end
    eval([cmd_savesbioproj ');']);
else
    sbioproject = m_sbioModels;
end

quasi_rand_stream = qrandstream( scramble( haltonset( length(m_sbioModels.m1.variant(1).content), 'Skip', 1e3, 'Leap',1e2), 'RR2' ) );
hqem_result_data( global_search_option.EnsembleRunNumber, 1 ) = struct( 'SamplParameter',[], 'SamplResNorm',[], 'OptParameter',[], 'OptResNorm',[] );
nextClearMemory = global_search_option.ClearMemoryStep;
while quasi_rand_stream.State < global_search_option.EnsembleRunNumber
    if quasi_rand_stream.State > nextClearMemory
        matlabpool('close');
        matlabpool('open',global_search_option.MatlabPoolSize);
        nextClearMemory = nextClearMemory + global_search_option.ClearMemoryStep;
    end
    variant_sampling_value = jz_sbioEnsemParamSampl( global_search_option.LowerBound, global_search_option.UpperBound, global_search_option.QuasiIterPopulation, quasi_rand_stream );
    hqem_result_data( (quasi_rand_stream.State - global_search_option.QuasiIterPopulation):(quasi_rand_stream.State - 1),1 ) = jz_sbioEnsemMultiModelOptim( ...
        sbioproject, fitting_value, variant_sampling_value, global_search_option.LowerBound, global_search_option.UpperBound, local_search_option, global_search_option.UseParallel );
    save( global_search_option.IntermediateFileName, 'm_sbioModels', 'global_search_option', 'local_search_option', 'fitting_value', 'hqem_result_data', 'quasi_rand_stream' );
end

hqem_parameter = zeros(global_search_option.EnsembleRunNumber,size(variant_sampling_value,2));
hqem_error = zeros(global_search_option.EnsembleRunNumber,1);
for t_n = 1:global_search_option.EnsembleRunNumber
    hqem_parameter(t_n,:) = hqem_result_data(t_n).OptParameter;
    hqem_error(t_n) = hqem_result_data(t_n).OptResNorm;
end
[hqem_error,hqem_sort_idx] = sort(hqem_error,'ascend');
hqem_parameter = hqem_parameter(hqem_sort_idx,:);

disp '-----------------------------------------------------------';
display([' HQEM caculation have been completed. Fitting results are saved in ' global_search_option.IntermediateFileName]);
if global_search_option.MatlabPoolSize > 1
    matlabpool('close');
    delete([sbioproject '.sbproj'])
end

end
